import asyncio
import json
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Mapping

import requests
import requests.adapters
from requests import Response
import logging
import ssl
from urllib.parse import quote, urlencode
from urllib3 import Retry
from tinybird.syncasync import sync_to_async


HOST = 'https://api.tinybird.co'
LIMIT_RETRIES = 10
LAST_PARTITION = 'last_partition'
ALL_PARTITIONS = 'all_partitions'
MAX_GET_LENGTH = 4096


class AuthException(Exception):
    pass


class AuthNoTokenException(AuthException):
    pass


class DoesNotExistException(Exception):
    pass


class CanNotBeDeletedException(Exception):
    pass


class OperationCanNotBePerformed(Exception):
    pass


class TimeoutException(Exception):
    pass


class ReachRetryLimit(Exception):
    pass


class ConnectorNothingToLoad(Exception):
    pass


class JobException(Exception):
    pass


def connector_equals(connector, datafile_params):
    if not connector:
        return False
    if connector['name'] == datafile_params['kafka_connection_name']:
        return True
    return False


def parse_error_response(response: Response) -> str:
    try:
        content: Dict = response.json()
        if content.get('error', None):
            error = content['error']
            if content.get('errors', None):
                error += f' -> errors: {content.get("errors")}'
        else:
            error = json.dumps(response, indent=4)
        return error
    except json.decoder.JSONDecodeError:
        return f"Server error, cannot parse response. {response.content.decode('utf-8')}"


class TinyB(object):

    def __init__(self, token: str, host: str = HOST, version: Optional[str] = None, disable_ssl_checks: bool = False):
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        self.token = token
        self.host = host
        self.version = version
        self.disable_ssl_checks = disable_ssl_checks

    async def _req(self, endpoint: str, data=None, files=None, method: str = 'GET', retries: int = 1, **kwargs):  # noqa: C901
        url = f"{self.host.strip('/')}/{endpoint.strip('/')}"

        if self.token:
            url += ('&' if '?' in endpoint else '?') + 'token=' + self.token
        if self.version:
            url += ('&' if '?' in url else '?') + 'cli_version=' + quote(self.version)

        verify_ssl = not self.disable_ssl_checks
        try:
            with requests.Session() as session:
                retry = Retry(
                    total=LIMIT_RETRIES,
                    status_forcelist=[429] if method in ('POST', 'PUT', 'DELETE') else [504, 502, 598, 599, 429],
                    allowed_methods=frozenset({method})
                )
                session.mount('https://', requests.adapters.HTTPAdapter(max_retries=retry))  # type: ignore
                session.mount('http://', requests.adapters.HTTPAdapter(max_retries=retry))  # type: ignore
                if method == 'POST':
                    if files:
                        response = await sync_to_async(session.post, thread_sensitive=False)(url, files=files, verify=verify_ssl, **kwargs)  # type: ignore
                    else:
                        response = await sync_to_async(session.post, thread_sensitive=False)(url, data=data, verify=verify_ssl, **kwargs)  # type: ignore
                elif method == 'PUT':
                    response = await sync_to_async(session.put, thread_sensitive=False)(url, data=data, verify=verify_ssl, **kwargs)  # type: ignore
                elif method == 'DELETE':
                    response = await sync_to_async(session.delete, thread_sensitive=False)(url, data=data, verify=verify_ssl, **kwargs)  # type: ignore
                else:
                    response = await sync_to_async(session.get, thread_sensitive=False)(url, verify=verify_ssl, **kwargs)  # type: ignore

        except Exception as e:
            raise e

        logging.debug("== server response ==")
        logging.debug(response.content)
        logging.debug("==      end        ==")

        if response.status_code == 403:
            error = parse_error_response(response)
            if not self.token:
                raise AuthNoTokenException(f"Forbidden: {error}")
            raise AuthException(f"Forbidden: {error}")
        if response.status_code == 204 or response.status_code == 205:
            return None
        if response.status_code == 404:
            error = parse_error_response(response)
            raise DoesNotExistException(error)
        if response.status_code == 400:
            error = parse_error_response(response)
            raise OperationCanNotBePerformed(error)
        if response.status_code == 409:
            error = parse_error_response(response)
            raise CanNotBeDeletedException(error)
        if response.status_code == 599:
            raise TimeoutException("timeout")
        if 'Content-Type' in response.headers and (response.headers['Content-Type'] == 'text/plain' or 'text/csv' in response.headers['Content-Type']):
            return response.content.decode('utf-8')
        if response.status_code >= 400 and response.status_code not in [400, 403, 404, 409, 429]:
            error = parse_error_response(response)
            raise Exception(error)
        if response.content:
            try:
                return response.json()
            except json.decoder.JSONDecodeError:
                raise Exception(f"Server error, cannot parse response. {response.content.decode('utf-8')}")

        return response

    async def tokens(self):
        tokens = await self._req("/v0/tokens")
        return tokens['tokens']

    async def starterkits(self) -> List[Dict[str, Any]]:
        data = await self._req("/v0/templates")
        return data.get('templates', [])

    async def get_token_by_name(self, name: str):
        tokens = await self.tokens()
        for tk in tokens:
            if tk['name'] == name:
                return tk
        return None

    async def create_token(self, name: str, scope: str, origin_code: Optional[str], origin_resource_name_or_id: Optional[str]):
        origin = origin_code or "C"  # == Origins.CUSTOM  if none specified
        return await self._req(f"/v0/tokens?name={name}&scope={scope}&origin={origin}&resource_id={origin_resource_name_or_id}", method='POST', data='')

    async def alter_tokens(self, name: str, scopes: List[str]):
        scopes_url: str = '&'.join([f'scope={scope}' for scope in scopes])
        return await self._req(f"/v0/tokens/{name}?{scopes_url}", method='PUT', data='')

    async def datasources(self, branch: Optional[str] = None, used_by: bool = False):
        params = {
            'attrs': 'used_by' if used_by else '',
        }
        response = await self._req(f"/v0/datasources?{urlencode(params)}")
        ds = response['datasources']

        if branch:
            ds = [x for x in ds if x['name'].startswith(branch)]
        return ds

    async def connections(self, connector: Optional[str] = None):
        response = await self._req('/v0/connectors')
        connectors = response['connectors']
        bigquery_connection = await self.bigquery_connection()
        connectors = connectors + [bigquery_connection] if bigquery_connection else connectors
        if connector:
            return [{
                'id': c['id'],
                'service': c['service'],
                'name': c['name'],
                'connected_datasources': len(c['linkers']),
                **c['settings']
            } for c in connectors if c['service'] == connector]
        return [{
            'id': c['id'],
            'service': c['service'],
            'name': c['name'],
            'connected_datasources': len(c['linkers']),
            **c['settings']
        } for c in connectors]

    async def bigquery_connection(self):
        bigquery_resources = await self.list_gcp_resources()
        if len(bigquery_resources) == 0:
            return None

        gcp_account_details: Dict[str, Any] = await self.get_gcp_service_account_details()
        datasources = await self.datasources()
        bigquery_datasources = [ds['name'] for ds in datasources if ds['type'] == 'bigquery']
        return {
            'id': gcp_account_details['account'].split('@')[0],
            'service': 'bigquery',
            'name': 'bigquery',
            'linkers': bigquery_datasources,
            'settings': gcp_account_details
        }

    async def get_datasource(self, ds_name: str, used_by: bool = False) -> Dict[str, Any]:
        params = {
            'attrs': 'used_by' if used_by else '',
        }
        return await self._req(f"/v0/datasources/{ds_name}?{urlencode(params)}")

    async def alter_datasource(self, ds_name: str, new_schema: Optional[str] = None, description: Optional[str] = None, ttl: Optional[str] = None, dry_run: bool = False):
        params = {'dry': 'true' if dry_run else 'false'}
        if new_schema:
            params.update({'schema': new_schema})
        if description:
            params.update({'description': description})
        if ttl:
            params.update({'ttl': ttl})
        res = await self._req(f'/v0/datasources/{ds_name}/alter?{urlencode(params)}', method='POST', data=b'')

        if 'Error' in res:
            raise Exception(res['error'])

        return res

    async def pipe_file(self, pipe: str):
        return await self._req(f"/v0/pipes/{pipe}.pipe")

    async def datasource_file(self, datasource: str):
        try:
            return await self._req(f"/v0/datasources/{datasource}.datasource")
        except DoesNotExistException:
            raise Exception(f"Data Source {datasource} not found.")

    async def datasource_analyze(self, url):
        params = {
            'url': url
        }
        return await self._req(f"/v0/analyze?{urlencode(params)}", method='POST', data='')

    async def datasource_analyze_file(self, data):
        return await self._req("/v0/analyze", method='POST', data=data)

    async def datasource_create_from_definition(self, parameter_definition: Dict[str, str]):
        return await self._req("/v0/datasources", method='POST', data=parameter_definition)

    async def datasource_create_from_url(self, table_name: str, url: str, mode: str = 'create', status_callback=None, sql_condition: Optional[str] = None, format: str = 'csv', replace_options: Optional[Set[str]] = None):
        params = {
            'name': table_name,
            'url': url,
            'mode': mode,
            'debug': 'blocks_block_log',
            'format': format
        }

        if sql_condition:
            params['replace_condition'] = sql_condition
        if replace_options:
            for option in list(replace_options):
                params[option] = 'true'

        req_url = f"/v0/datasources?{urlencode(params, safe='')}"
        res = await self._req(req_url, method='POST', data=b'')

        if 'error' in res:
            raise Exception(res['error'])

        return await self.wait_for_job(res['id'], status_callback, backoff_multiplier=1.5, maximum_backoff_seconds=20)

    async def datasource_delete(self, datasource_name: str, force: bool = False, dry_run: bool = False):
        params = {
            'force': 'true' if force else 'false',
            'dry_run': 'true' if dry_run else 'false'
        }
        return await self._req(f"/v0/datasources/{datasource_name}?{urlencode(params)}", method='DELETE')

    async def datasource_append_data(self, datasource_name: str, file: Path, mode: str = 'append', status_callback=None, sql_condition: Optional[str] = None, format: str = 'csv', replace_options: Optional[Set[str]] = None):
        params = {
            'name': datasource_name,
            'mode': mode,
            'format': format,
            'debug': 'blocks_block_log'
        }

        if sql_condition:
            params['replace_condition'] = sql_condition
        if replace_options:
            for option in list(replace_options):
                params[option] = 'true'

        with open(file, 'rb') as content:
            if format == 'csv':
                files = {'csv': ('csv', content)}
            else:
                files = {'ndjson': ('ndjson', content)}

            res = await self._req(
                f"v0/datasources?{urlencode(params, safe='')}",
                files=files,
                method='POST',
            )
            if status_callback:
                status_callback(res)

            return res

    async def datasource_truncate(self, datasource_name: str):
        return await self._req(f"/v0/datasources/{datasource_name}/truncate", method='POST', data='')

    async def datasource_delete_rows(self, datasource_name: str, delete_condition: str):
        params = {
            "delete_condition": delete_condition
        }
        return await self._req(f'/v0/datasources/{datasource_name}/delete', method='POST', data=params)

    async def datasource_dependencies(self, no_deps: bool, match: str, pipe: str, datasource: str, check_for_partial_replace: bool, recursive: bool):
        params = {
            "no_deps": 'true' if no_deps else 'false',
            "check_for_partial_replace": 'true' if check_for_partial_replace else 'false',
            "recursive": 'true' if recursive else 'false'
        }
        if match:
            params['match'] = match
        if pipe:
            params['pipe'] = pipe
        if datasource:
            params['datasource'] = datasource

        return await self._req(f'/v0/dependencies?{urlencode(params)}', timeout=60)

    async def datasource_share(self, datasource_id: str, current_workspace_id: str, destination_workspace_id: str):
        params = {
            'origin_workspace_id': current_workspace_id,
            'destination_workspace_id': destination_workspace_id
        }
        return await self._req(f"/v0/datasources/{datasource_id}/share", method='POST', data=params)

    async def datasource_unshare(self, datasource_id: str, current_workspace_id: str, destination_workspace_id: str):
        params = {
            'origin_workspace_id': current_workspace_id,
            'destination_workspace_id': destination_workspace_id
        }
        return await self._req(f"/v0/datasources/{datasource_id}/share", method='DELETE', data=params)

    async def analyze_pipe_node(self, pipe_name: str, node: Dict[str, Any], dry_run: str = 'false', datasource_name: str = None):
        params = {**{
            'include_datafile': 'true',
            'dry_run': dry_run
        }, **node.get('params', node)}
        node_name = node['params']['name'] if node.get('params', None) else node['name']
        if datasource_name:
            params['datasource'] = datasource_name
        response = await self._req(f"/v0/pipes/{pipe_name}/nodes/{node_name}/analysis?{urlencode(params)}")
        return response

    async def populate_node(self, pipe_name: str, node_name: str, populate_subset: bool = False, populate_condition: str = None, truncate: bool = True, unlink_on_populate_error: bool = False):
        params: Dict[str, Any] = {
            'truncate': 'true' if truncate else 'false',
            'unlink_on_populate_error': 'true' if unlink_on_populate_error else 'false'
        }
        if populate_subset:
            params.update({'populate_subset': populate_subset})
        if populate_condition:
            params.update({'populate_condition': populate_condition})
        response = await self._req(f"/v0/pipes/{pipe_name}/nodes/{node_name}/population?{urlencode(params)}", method='POST')
        return response

    async def pipes(self, branch=None, dependencies: bool = False, node_attrs=None, attrs=None):
        params = {
            'dependencies': 'true' if dependencies else 'false',
            'attrs': attrs if attrs else '',
            'node_attrs': node_attrs if node_attrs else '',
        }
        response = await self._req(f"/v0/pipes?{urlencode(params)}")
        pipes = response['pipes']
        if branch:
            pipes = [x for x in pipes if x['name'].startswith(branch)]
        return pipes

    async def pipe(self, pipe: str):
        return await self._req(f"/v0/pipes/{pipe}")

    async def pipe_data(self, pipe_name_or_uid: str, sql: Optional[str] = None, format: Optional[str] = 'json', params: Mapping[str, Any] = None):
        params = {**params} if params else {}
        if sql:
            params['q'] = sql

        return await self._req(f"/v0/pipes/{pipe_name_or_uid}.{format}?{urlencode(params)}")

    async def pipe_create(self, pipe_name: str, sql: str):
        return await self._req(f"/v0/pipes?name={pipe_name}&sql={quote(sql, safe='')}", method='POST', data=sql.encode())

    async def pipe_delete(self, pipe_name: str):
        return await self._req(f"/v0/pipes/{pipe_name}", method='DELETE')

    async def pipe_append_node(self, pipe_name_or_uid: str, sql: str):
        return await self._req(f"/v0/pipes/{pipe_name_or_uid}/nodes", method='POST', data=sql.encode())

    async def pipe_set_endpoint(self, pipe_name_or_uid: str, published_node_uid: str):
        return await self._req(f"/v0/pipes/{pipe_name_or_uid}/nodes/{published_node_uid}/endpoint", method='POST')

    async def pipe_remove_endpoint(self, pipe_name_or_uid: str, published_node_uid: str):
        return await self._req(f"/v0/pipes/{pipe_name_or_uid}/nodes/{published_node_uid}/endpoint", method='DELETE')

    async def query(self, sql: str, pipeline: str = None):
        params = {}
        if pipeline:
            params = {
                'pipeline': pipeline
            }

        if len(sql) > MAX_GET_LENGTH:
            return await self._req(f"/v0/sql?{urlencode(params)}", data=sql, method='POST')
        else:
            return await self._req(f"/v0/sql?q={quote(sql, safe='')}&{urlencode(params)}")

    async def jobs(self, status=None):
        jobs = (await self._req("/v0/jobs"))['jobs']
        if status:
            status = [status] if isinstance(status, str) else status
            jobs = [j for j in jobs if j['status'] in status]
        return jobs

    async def job(self, job_id: str):
        return await self._req(f"/v0/jobs/{job_id}")

    async def job_cancel(self, job_id: str):
        return await self._req(f"/v0/jobs/{job_id}/cancel", method='POST', data=b'')

    async def user_workspaces(self):
        return await self._req("/v0/user/workspaces/?with_branches=false")

    async def user_workspaces_and_branches(self):
        return await self._req("/v0/user/workspaces/?with_branches=true")

    async def user_workspace_branches(self):
        return await self._req("/v0/user/workspaces/?with_branches=true&only_branches=true")

    async def branches(self):
        return await self._req("/v0/branches")

    async def create_workspace(self, name: str, template: Optional[str]):
        url = f"/v0/workspaces?name={name}"
        if template:
            url += f'&starter_kit={template}'
        return await self._req(url, method='POST', data=b'')

    async def create_workspace_branch(self, branch_name: str, last_partition: Optional[bool], all: Optional[bool]):
        params = {
            'name': branch_name,
        }
        if last_partition:
            params['data'] = LAST_PARTITION

        if all:
            params['data'] = ALL_PARTITIONS
        return await self._req(f"/v0/branches?{urlencode(params)}", method='POST',
                               data=b'')

    async def merge_workspace_branch(self, branch_id: str):
        return await self._req(f"/v0/branches/{branch_id}/deployment", method='POST', data=b'')

    async def branch_workspace_data(self, workspace_id: str, last_partition: Optional[str], all: Optional[str]):
        params = {}
        if last_partition:
            params['mode'] = LAST_PARTITION

        if all:
            params['mode'] = ALL_PARTITIONS
        url = f"/v0/branches/{workspace_id}/data?{urlencode(params)}"
        return await self._req(url, method='POST', data=b'')

    async def branch_regression_tests(self, branch_id: str, pipes: Optional[List[str]], test_type: str, only_response_times: bool,
                                      failfast: bool, ignore_order: bool, validate_processed_bytes: bool,
                                      limit: Optional[int] = None, samples_by_params: Optional[int] = None, matches: Optional[List[str]] = None,
                                      params: Optional[List[Dict[str, Any]]] = None):
        regression_commands: List[Dict[str, Any]] = []
        test: Dict[str, Any] = {test_type: {'config': {'ignore_order': ignore_order,
                                                       'failfast': failfast,
                                                       'only_response_times': only_response_times,
                                                       'processed_bytes': validate_processed_bytes}}}
        if limit is not None:
            test[test_type].update({'limit': limit})
        if samples_by_params is not None:
            test[test_type].update({'samples_by_params': samples_by_params})
        if matches is not None:
            test[test_type].update({'matches': matches})
        if params is not None:
            test[test_type].update({'params': params})

        if not pipes:
            pipes = ["*"]
        for pipe in pipes:
            regression_commands.append({'pipe': "*" if pipe is None else pipe,
                                        'tests': [test]})
        data = json.dumps(regression_commands)
        url = f"/v0/branches/{branch_id}/regression"
        return await self._req(url, method='POST', data=data, headers={'Content-Type': 'application/json'})

    async def branch_regression_tests_file(self, branch_id: str, regression_commands: List[Dict[str, Any]]):
        data = json.dumps(regression_commands)
        url = f"/v0/branches/{branch_id}/regression"
        return await self._req(url, method='POST', data=data, headers={'Content-Type': 'application/json'})

    async def delete_workspace(self, id: str, hard_delete_confirmation: Optional[str]):
        data = {
            'confirmation': hard_delete_confirmation
        }
        return await self._req(f"/v0/workspaces/{id}", data, method='DELETE')

    async def delete_branch(self, id: str):
        return await self._req(f"/v0/branches/{id}", method='DELETE')

    async def add_users_to_workspace(self, workspace: Dict[str, Any], users_emails: List[str]):
        users = ','.join(users_emails)
        return await self._req(f"/v0/workspaces/{workspace['id']}/users/", method='PUT', data={'operation': 'add', 'users': users})

    async def remove_users_from_workspace(self, workspace: Dict[str, Any], users_emails: List[str]):
        users = ','.join(users_emails)
        return await self._req(f"/v0/workspaces/{workspace['id']}/users/", method='PUT', data={'operation': 'remove', 'users': users})

    async def set_role_for_users_in_workspace(self, workspace: Dict[str, Any], users_emails: List[str], role: str):
        users = ','.join(users_emails)
        return await self._req(f"/v0/workspaces/{workspace['id']}/users/", method='PUT', data={'operation': 'change_role', 'users': users, 'new_role': role})

    async def workspace(self, workspace_id: str, with_token: bool = False):
        params = {
            'with_token': 'true' if with_token else 'false'
        }
        return await self._req(f"/v0/workspaces/{workspace_id}?{urlencode(params)}")

    async def workspace_info(self):
        return await self._req("/v0/workspace")

    async def wait_for_job(
            self,
            job_id: str,
            status_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
            backoff_seconds: float = 2.0,
            backoff_multiplier: float = 1,
            maximum_backoff_seconds: float = 2.0):
        done: bool = False
        while not done:
            res: Dict[str, Any] = await self._req("/v0/jobs/" + job_id + '?debug=blocks,block_log')

            if res['status'] == 'error':
                error_message = 'There has been an error'
                if not isinstance(res.get('error', True), bool):
                    error_message = str(res['error'])
                if 'errors' in res:
                    error_message += f": {res['errors']}"
                raise JobException(error_message)

            if res['status'] == 'cancelled':
                raise JobException('Job has been cancelled')

            done = res['status'] == 'done'

            if status_callback:
                status_callback(res)

            if not done:
                backoff_seconds = min(backoff_seconds * backoff_multiplier, maximum_backoff_seconds)
                await asyncio.sleep(backoff_seconds)

        return res

    async def datasource_kafka_connect(self, connection_id, datasource_name, topic, group, auto_offset_reset):
        return await self._req(f"/v0/datasources?connector={connection_id}&name={datasource_name}&"
                               f"kafka_topic={topic}&kafka_group_id={group}&kafka_auto_offset_reset={auto_offset_reset}",
                               method='POST', data=b'')

    async def connection_create_kafka(
        self,
        kafka_bootstrap_servers,
        kafka_key,
        kafka_secret,
        kafka_connection_name,
        kafka_auto_offset_reset=None,
        kafka_schema_registry_url=None,
        kafka_sasl_mechanism="PLAIN"
    ):

        params = {
            'service': 'kafka',
            'kafka_security_protocol': 'SASL_SSL',
            'kafka_sasl_mechanism': kafka_sasl_mechanism,
            'kafka_bootstrap_servers': kafka_bootstrap_servers,
            'kafka_sasl_plain_username': kafka_key,
            'kafka_sasl_plain_password': kafka_secret,
            'name': kafka_connection_name
        }

        if kafka_schema_registry_url:
            params['kafka_schema_registry_url'] = kafka_schema_registry_url
        if kafka_auto_offset_reset:
            params['kafka_auto_offset_reset'] = kafka_auto_offset_reset

        connection_params = {
            key: value
            for key, value in params.items() if value is not None
        }

        return await self._req(f"/v0/connectors?{urlencode(connection_params)}", method='POST', data='')

    async def kafka_list_topics(self, connection_id: str, timeout=5):
        resp = await self._req(f"/v0/connectors/{connection_id}/preview?preview_activity=false",
                               connect_timeout=timeout, request_timeout=timeout)
        return [x['topic'] for x in resp['preview']]

    async def get_gcp_service_account_details(self) -> Dict[str, Any]:
        return await self._req("/v0/datasources-bigquery-credentials")

    async def list_gcp_resources(self) -> List[Dict[str, Any]]:
        try:
            resources = await self._req("/v0/connections/bigquery")
            if not resources:
                return []

            return resources.get('items', [])
        except Exception:
            return []

    async def check_gcp_read_permissions(self) -> bool:
        """Returns `True` if our service account (see `TinyB::get_gcp_service_account_details()`)
           has the proper permissions in GCP.

           Here we assume that we have permissions if we can list resources but currently this
           logic is wrong under some circumstances.

           See https://gitlab.com/tinybird/analytics/-/issues/6485.
        """
        try:
            items = await self.list_gcp_resources()
            if not items:
                return False
            return len(items) > 0
        except Exception:
            return False

    async def connector_delete(self, connection_id):
        return await self._req(f"/v0/connectors/{connection_id}", method='DELETE')

    @staticmethod
    def _sql_get_used_tables_local(sql: str, raising: bool = False) -> List[str]:
        from tinybird.sql_toolset import sql_get_used_tables
        tables = sql_get_used_tables(sql, raising, table_functions=False)
        return [t[1] if t[0] == "" else f"{t[0]}.{t[1]}" for t in tables]

    async def _sql_get_used_tables_remote(self, sql: str, raising: bool = False) -> List[str]:
        params = {
            'q': sql,
            'raising': 'true' if raising else 'false',
            'table_functions': 'false'
        }
        result = await self._req('/v0/sql_tables', data=params, method='POST')
        return [t[1] if t[0] == "" else f"{t[0]}.{t[1]}" for t in result['tables']]

    # Get used tables from a query. Does not include table functions
    async def sql_get_used_tables(self, sql: str, raising: bool = False) -> List[str]:
        try:
            return self._sql_get_used_tables_local(sql, raising)
        except ModuleNotFoundError:
            return await self._sql_get_used_tables_remote(sql, raising)

    @staticmethod
    def _replace_tables_local(q: str, replacements):
        from tinybird.sql_toolset import replace_tables, replacements_to_tuples
        return replace_tables(q, replacements_to_tuples(replacements))

    async def _replace_tables_remote(self, q: str, replacements):
        params = {
            'q': q,
            'replacements': json.dumps({k[1] if isinstance(k, tuple) else k: v
                                        for k, v in replacements.items()})
        }
        result = await self._req('/v0/sql_replace', data=params, method='POST')
        return result['query']

    async def replace_tables(self, q: str, replacements):
        try:
            return self._replace_tables_local(q, replacements)
        except ModuleNotFoundError:
            return await self._replace_tables_remote(q, replacements)

    async def get_connection(self, **kwargs):
        result = await self._req('/v0/connectors')
        return next((connector for connector in result['connectors'] if connector_equals(connector, kwargs)), None)

    async def regions(self):
        return await self._req('/v0/regions')

    async def datasource_query_copy(self, datasource_name: str, sql_query: str):
        return await self._req(f'/v0/sql_copy?copy_to={datasource_name}',
                               data=sql_query,
                               method='POST')



import asyncio
import tornado

from tests.views.base_test import BaseTest
from tests_e2e.aux import poll_async
from tinybird.async_redis import async_redis_writer
from tinybird.redis_client import get_redis_args

from datetime import datetime, timezone
from io import StringIO
from unittest.mock import AsyncMock

import aiohttp
from tinybird.user import Users
from tinybird.token_scope import scopes
from tinybird.ingestion_observer import INGESTION_LAGGY_WINDOW, INGESTION_OBSERVER_POLL_SECONDS, IngestionObserver
from .conftest import is_main_process
import pytest


@pytest.mark.serial  # Better to test this serially
@pytest.mark.skipif(not is_main_process(), reason="Serial test")
class TestIncident(BaseTest):
    @tornado.testing.gen_test
    async def test_ingest_incident_ndjson_url_failed(self):
        ds_id = None

        async def provoke_incident():
            nonlocal ds_id
            u = Users.get_by_id(self.WORKSPACE_ID)
            token = Users.get_token_for_scope(u, scopes.ADMIN)
            headers = {'Authorization': f"Bearer {token}"}
            ds_name = 'test_ingest_incident_ndjson_url_failed'
            async with aiohttp.request('POST', self.get_url('/v0/datasources'),
                                       headers=headers,
                                       params={
                                       'mode': 'create',
                                       'name': ds_name,
                                       'format': 'ndjson',
                                       'schema': "date DateTime('UTC') `json:$.date`, city String `json:$.extra_data.city`",
                                       'url': self.get_url('/not_exists.ndjson'),
                                       }) as r:
                self.assertEqual(r.status, 200, await r.json())
                ds_id = (await r.json())['job']['datasource']['id']

        async def check_mailgun_mock(mailgun_mock):
            nonlocal ds_id
            assert mailgun_mock.send_notification_on_ingestion_incident.call_count == 1
            call_args = mailgun_mock.send_notification_on_ingestion_incident.call_args.args
            assert '@example.com' in call_args[0][0]
            assert call_args[2].id == ds_id
            assert call_args[3]['status'] == 'new'
            assert '404' in call_args[3]['errors'][0]
            assert call_args[3]['next_timestamp'] == 0

        await self._test_incident(provoke_incident, check_mailgun_mock)

    @tornado.testing.gen_test
    async def test_ingest_quarantine_incident(self):
        ds_id = None

        async def provoke_incident():
            nonlocal ds_id
            u = Users.get_by_id(self.WORKSPACE_ID)
            token = Users.get_token_for_scope(u, scopes.ADMIN)
            headers = {'Authorization': f"Bearer {token}"}
            ds_name = 'test_ingest_quarantine_incident'
            async with aiohttp.request('POST', self.get_url('/v0/datasources'),
                                       headers=headers,
                                       params={
                                       'mode': 'create',
                                       'name': ds_name,
                                       'schema': 'col_a Int32,col_b Int32,col_c Int8',
                                       }) as r:
                self.assertEqual(r.status, 200, await r.json())
                ds_id = (await r.json())['datasource']['id']
            params = {
                'mode': 'append',
                'token': token,
                'name': ds_name,
            }
            data = StringIO('a,b,3\n4,d,6\n7,h,j')
            async with aiohttp.request('POST', self.get_url('/v0/datasources'),
                                       headers=headers,
                                       params=params,
                                       data=data) as r:
                assert r.status == 200

        async def check_mailgun_mock(mailgun_mock):
            nonlocal ds_id
            assert mailgun_mock.send_notification_on_quarantine_incident.call_count == 1
            call_args = mailgun_mock.send_notification_on_quarantine_incident.call_args.args
            assert '@example.com' in call_args[0][0]
            assert call_args[2].id == ds_id
            assert call_args[3]['status'] == 'new'
            assert call_args[3]['imports'] == 1
            assert call_args[3]['rows'] == 3
            assert call_args[3]['next_timestamp'] == 0

        await self._test_incident(provoke_incident, check_mailgun_mock)

    async def _test_incident(self, provoke_incident, check_mailgun_mock):
        redis_args = get_redis_args()
        async_redis_writer.init({
            'redis_host': redis_args['host'],
            'redis_db': redis_args['db']})
        mailgun_mock = AsyncMock()
        ingestion_observer = IngestionObserver(mailgun_mock)

        async def f():
            last_check = int(datetime.now(timezone.utc).timestamp()) - 2 * INGESTION_OBSERVER_POLL_SECONDS
            return True, last_check

        ingestion_observer._should_check_for_incidents = f
        await ingestion_observer.run()
        try:
            await provoke_incident()
            await asyncio.sleep(INGESTION_LAGGY_WINDOW)

            async def f():
                await check_mailgun_mock(mailgun_mock)

            await poll_async(f)
        finally:
            await ingestion_observer.terminate()
            await async_redis_writer.reset()



import threading
from collections import defaultdict

import unittest
from unittest.mock import patch
import uuid
from dataclasses import dataclass

import tornado.testing
from tornado.testing import AsyncTestCase

from tinybird.job import Job
from tinybird.model import retry_transaction_in_case_of_concurrent_edition_error_sync, ConcurrentEditionException, \
    retry_transaction_in_case_of_concurrent_edition_error_async, RedisModel, ModelLRUCache
from tests.test_jobs import FakeTestJob
from tinybird.user import User, UserAccount
from tinybird.redis_client import get_redis_client


class FailsBeforeSuccess:
    def __init__(self, fails):
        self.fails = fails


class TestRetryInCaseOfConcurrentEditionError(AsyncTestCase):
    def setUp(self):
        super().setUp()
        workspace_name = f'testing_model_{uuid.uuid4().hex}'
        user_email = f'{workspace_name}@example.com'

        self.user = UserAccount.register(user_email, 'pass')
        self.workspace = User.register(workspace_name, admin=self.user.id)
        self.workspace.database = 'default'
        self.workspace.save()

    def tearDown(self):
        User._delete(self.workspace.id)
        super().tearDown()

    def test_correctly_finished_after_some_retries(self):
        job = FakeTestJob(self.workspace)
        job.save()

        @retry_transaction_in_case_of_concurrent_edition_error_sync(tries=10, delay=0.001, backoff=1)
        def semi_mocked_function_with_some_retries(fail_before_success):
            with Job.transaction(job.id) as j:
                if fail_before_success.fails != 0:
                    print(fail_before_success.fails)
                    fail_before_success.fails -= 1
                    j.save()
            return fail_before_success.fails

        remaining_fails = semi_mocked_function_with_some_retries(FailsBeforeSuccess(8))
        self.assertEqual(remaining_fails, 0)

    def test_raises_runtime_error_if_it_fails_more_than_the_configured_retries(self):
        job = FakeTestJob(self.workspace)
        job.save()

        @retry_transaction_in_case_of_concurrent_edition_error_sync(tries=3, delay=0.001, backoff=1)
        def semi_mocked_function_with_some_retries(fail_before_success):
            with Job.transaction(job.id) as j:
                if fail_before_success.fails != 0:
                    print(fail_before_success.fails)
                    fail_before_success.fails -= 1
                    j.save()

        with self.assertRaises(ConcurrentEditionException):
            semi_mocked_function_with_some_retries(FailsBeforeSuccess(5))

    @tornado.testing.gen_test
    async def test_correctly_finished_after_some_retries_async(self):
        job = FakeTestJob(self.workspace)
        job.save()

        @retry_transaction_in_case_of_concurrent_edition_error_async(tries=10, delay=0.001, backoff=1)
        async def semi_mocked_function_with_some_retries(fail_before_success):
            with Job.transaction(job.id) as j:
                if fail_before_success.fails != 0:
                    print(fail_before_success.fails)
                    fail_before_success.fails -= 1
                    j.save()
            return fail_before_success.fails

        remaining_fails = await semi_mocked_function_with_some_retries(FailsBeforeSuccess(8))
        self.assertEqual(remaining_fails, 0)

    @tornado.testing.gen_test
    async def test_raises_runtime_error_if_it_fails_more_than_the_configured_retries_async(self):
        job = FakeTestJob(self.workspace)
        job.save()

        @retry_transaction_in_case_of_concurrent_edition_error_async(tries=3, delay=0.001, backoff=1)
        async def semi_mocked_function_with_some_retries(job, fail_before_success):
            with Job.transaction(job.id) as j:
                if fail_before_success.fails != 0:
                    print(fail_before_success.fails)
                    fail_before_success.fails -= 1
                    j.save()
            return fail_before_success.fails

        with self.assertRaises(ConcurrentEditionException):
            await semi_mocked_function_with_some_retries(job, FailsBeforeSuccess(5))


class ParentClass(RedisModel):
    __namespace__ = 'parentclass'

    __props__ = [
        'id',
        'name'
    ]

    def __init__(self, **config):
        self.id = None
        self.name = None

        super().__init__(**config)


class ChildClass(RedisModel):
    __namespace__ = 'childclass'
    __owner__ = 'parent_id'

    __props__ = [
        'id',
        'name',
        'parent_id'
    ]

    def __init__(self, **config):
        self.id = None
        self.name = None
        self.parent_id = None

        super().__init__(**config)


class ChildClassOwnerLimited(ChildClass):
    __owner_max_children__ = 100


class TestModelOwnership(unittest.TestCase):
    def setUp(self):
        super().setUp()
        redis_client = get_redis_client()
        ParentClass.config(redis_client)
        ChildClass.config(redis_client)
        ChildClassOwnerLimited.config(redis_client)

    def create_parent(self, name):
        uid = str(uuid.uuid4())
        config = {
            'id': uid,
            'name': name
        }

        parent = ParentClass(**config)
        parent.save()
        return parent

    def create_child(self, parent_id, name, child_cls=ChildClass):
        uid = str(uuid.uuid4())
        config = {
            'id': uid,
            'parent_id': parent_id,
            'name': name
        }

        child = child_cls(**config)
        child.save()
        return child

    def test_simple_ownership(self):
        parent = self.create_parent('parent')
        child = self.create_child(parent.id, 'child_1')
        self.assertTrue(ChildClass.is_owned_by(child.id, parent.id))

    def test_model_with_owner_is_deleted_when_the_owner_is_deleted(self):
        parent = self.create_parent('parent')
        child = self.create_child(parent.id, 'child_1')

        children = ChildClass.get_all_by_owner(parent.id)

        self.assertEqual(children[0].id, child.id)

        ChildClass._delete(child.id)
        children = ChildClass.get_all_by_owner(parent.id)

        self.assertEqual(len(children), 0)

    @patch('random.randint', lambda a, b: RedisModel.OWNER_SET_REMOVE_EVERY_X_ADD)
    def test_get_all_by_owner_lazy_deletion_not_removing_by_default(self):
        redis_client = get_redis_client()

        parent = self.create_parent('parent')
        parent_owner_key = f"{ChildClass.__namespace__}:owner:{parent.id}"

        HOW_MANY_CHILDREN = 125

        def child_name(i):
            return f"child_{i}"

        for i in range(HOW_MANY_CHILDREN):
            self.create_child(parent.id, child_name(i))

        HOW_MANY_REQUESTED = 20

        # without any other action
        # we have as many children in the set as we created
        # even if the probabilistic deletion matches
        children_count = redis_client.zcard(parent_owner_key)
        self.assertEqual(children_count, HOW_MANY_CHILDREN)

        # We should return up to the limit
        children = ChildClass.get_all_by_owner(parent.id, limit=HOW_MANY_REQUESTED)
        self.assertEqual(len(children), HOW_MANY_REQUESTED)
        for i in range(HOW_MANY_REQUESTED):
            self.assertEqual(children[i].name, child_name(HOW_MANY_CHILDREN - i - 1))

        # But we should not remove anything from the set
        children_count = redis_client.zcard(parent_owner_key)
        self.assertEqual(children_count, HOW_MANY_CHILDREN)

        # We should keep them sorted by last creation
        children = ChildClass.get_all_by_owner(parent.id, limit=HOW_MANY_CHILDREN)
        for i in range(HOW_MANY_CHILDREN):
            self.assertEqual(children[i].name, child_name(HOW_MANY_CHILDREN - i - 1))

    @patch('random.randint', lambda a, b: 1)
    def test_get_all_by_owner_lazy_deletion(self):
        redis_client = get_redis_client()

        parent = self.create_parent('parent')
        parent_owner_key = f"{ChildClassOwnerLimited.__namespace__}:owner:{parent.id}"

        HOW_MANY_CHILDREN = 125

        def child_name(i):
            return f"child_{i}"

        for i in range(HOW_MANY_CHILDREN):
            self.create_child(parent.id, child_name(i), child_cls=ChildClassOwnerLimited)

        # without any other action and the probabilistic deletion,
        # we have as many children in the set as we created
        children_count = redis_client.zcard(parent_owner_key)
        self.assertEqual(children_count, HOW_MANY_CHILDREN)

        # We should not return more children than the limit
        children = ChildClassOwnerLimited.get_all_by_owner(parent.id, limit=HOW_MANY_CHILDREN)
        self.assertEqual(len(children), ChildClassOwnerLimited.__owner_max_children__)

        # We should have lazily deleted all but 100 children which is the limit
        children_count = redis_client.zcard(parent_owner_key)
        self.assertEqual(children_count, ChildClassOwnerLimited.__owner_max_children__)

        # We should also keep the most recent children
        children = ChildClassOwnerLimited.get_all_by_owner(parent.id, limit=ChildClassOwnerLimited.__owner_max_children__)
        for i in range(ChildClassOwnerLimited.__owner_max_children__):
            self.assertEqual(children[i].name, child_name(HOW_MANY_CHILDREN - i - 1))

    @patch('random.randint', lambda a, b: 1)
    def test_get_all_by_owner_lazy_deletion_with_smaller_limit(self):
        redis_client = get_redis_client()

        parent = self.create_parent('parent')
        parent_owner_key = f"{ChildClassOwnerLimited.__namespace__}:owner:{parent.id}"

        HOW_MANY_CHILDREN = 125

        def child_name(i):
            return f"child_{i}"

        for i in range(HOW_MANY_CHILDREN):
            self.create_child(parent.id, child_name(i), child_cls=ChildClassOwnerLimited)

        HOW_MANY_REQUESTED = 20

        # without any other action we have as many children in the set as we created
        children_count = redis_client.zcard(parent_owner_key)
        self.assertEqual(children_count, HOW_MANY_CHILDREN)

        # We should return up to the limit
        children = ChildClassOwnerLimited.get_all_by_owner(parent.id, limit=HOW_MANY_REQUESTED)
        self.assertEqual(len(children), HOW_MANY_REQUESTED)
        for i in range(HOW_MANY_REQUESTED):
            self.assertEqual(children[i].name, child_name(HOW_MANY_CHILDREN - i - 1))

        # We should have lazily deleted all but 100 children which is the limit
        children_count = redis_client.zcard(parent_owner_key)
        self.assertEqual(children_count, ChildClassOwnerLimited.__owner_max_children__)

        # We should also keep the most recent children
        children = ChildClassOwnerLimited.get_all_by_owner(parent.id, limit=ChildClassOwnerLimited.__owner_max_children__)
        for i in range(ChildClassOwnerLimited.__owner_max_children__):
            self.assertEqual(children[i].name, child_name(HOW_MANY_CHILDREN - i - 1))

    @patch('random.randint', lambda a, b: RedisModel.OWNER_SET_REMOVE_EVERY_X_ADD)
    def test_get_all_by_owner_probabilistic_deletion_on_add(self):
        redis_client = get_redis_client()

        parent = self.create_parent('parent')
        parent_owner_key = f"{ChildClassOwnerLimited.__namespace__}:owner:{parent.id}"

        HOW_MANY_CHILDREN = 125

        def child_name(i):
            return f"child_{i}"

        for i in range(HOW_MANY_CHILDREN):
            self.create_child(parent.id, child_name(i), child_cls=ChildClassOwnerLimited)

        # The probabilistic deletion should remove any extra children
        children_count = redis_client.zcard(parent_owner_key)
        self.assertEqual(children_count, ChildClassOwnerLimited.__owner_max_children__)

        # Still, we should keep the most recent children
        children = ChildClassOwnerLimited.get_all_by_owner(parent.id, limit=ChildClassOwnerLimited.__owner_max_children__)
        for i in range(ChildClassOwnerLimited.__owner_max_children__):
            self.assertEqual(children[i].name, child_name(HOW_MANY_CHILDREN - i - 1))


@dataclass
class ModelMigrationThreadConditions:
    cv_migration: threading.Condition
    is_first_thread: bool
    cv_transaction_finish: threading.Condition
    transaction_finished: bool


class TestModelMigrationDoesNotLeaveCorruptedData(unittest.TestCase):

    stored_thread_conditions = defaultdict(lambda: ModelMigrationThreadConditions(
        cv_migration=threading.Condition(),
        is_first_thread=True,
        cv_transaction_finish=threading.Condition(),
        transaction_finished=False
    ))

    # The first time this is called it is by the "slow thread", to simulate this we wait until the main transaction is
    # done The second time this is called it is by the fast thread, where we don't wait for anything and overtake
    # the slow thread
    @staticmethod
    def migration_with_wait(u):
        print(str(threading.get_ident()) + " Started migration")

        thread_conditions = TestModelMigrationDoesNotLeaveCorruptedData.stored_thread_conditions[u['id']]
        is_first = False
        with thread_conditions.cv_migration:
            if thread_conditions.is_first_thread:
                is_first = True
                thread_conditions.is_first_thread = False
                thread_conditions.cv_migration.notify()

        if is_first:
            with thread_conditions.cv_transaction_finish:
                thread_conditions.cv_transaction_finish.wait()

        u['value_to_check'] = "value_written_in_migration"
        print(str(threading.get_ident()) + " completed migration")
        return u

    # This is a slow thread transaction
    # It is executed first so it starts the migrations, but before it is done a second thread also starts the
    # transactions and it finishes first. When the first thread finishes, as there has been a concurrent edit, it
    # will get an exception and retry to read the object; as the object has already been migrated (thread #2) it
    # doesn't need to run migrations again
    def _execute_slow_transaction(self, example_id):
        print(str(threading.get_ident()) + " Started change in SLOW thread")

        migration_id = ExampleModelForMigrationChecks.get_by_id(example_id)
        self.assertEqual(migration_id.value_to_check, 'change_in_fast_transaction')
        print(str(threading.get_ident()) + " Completed change in SLOW thread")

    def test_modifications_to_user_returned_by_get_by_id_does_not_appears_as_modified_in_other_objects(self):

        example_id = f'example_id_{uuid.uuid4().hex}'

        example = ExampleModelForMigrationChecks(id=example_id)
        example.save()

        # The migration was not executed as the object is directly inserted in the cache with the .save() and is
        # returned as it is using the get_by_id or the transaction.
        # Cleaning the cache will force the retrieval of the Redis model so the migration will be executed.
        ExampleModelForMigrationChecks.__object_cache_by_id__ = defaultdict(lambda: ModelLRUCache(128))

        parallel_exec = threading.Thread(target=self._execute_slow_transaction, args=(example_id,))
        parallel_exec.start()

        thread_conditions = TestModelMigrationDoesNotLeaveCorruptedData.stored_thread_conditions[example_id]

        # Don't do anything until the slow thread has already started doing migrations
        with thread_conditions.cv_migration:
            thread_conditions.cv_migration.wait()

        print(str(threading.get_ident()) + " Starting transaction in FAST thread")
        with ExampleModelForMigrationChecks.transaction(example_id) as example:
            example.value_to_check = "change_in_fast_transaction"

        print(str(threading.get_ident()) + " Completed change in FAST thread")
        # Notify the slow thread that the fast thread has finished so it stops waiting (simulating it's slow)
        with thread_conditions.cv_transaction_finish:
            thread_conditions.transaction_finished = True
            thread_conditions.cv_transaction_finish.notify()

        parallel_exec.join()
        example = ExampleModelForMigrationChecks.get_by_id(example_id)

        self.assertEqual(example.value_to_check, 'change_in_fast_transaction')


class ExampleModelForMigrationChecks(RedisModel):
    __namespace__ = 'examplemodel'
    __props__ = ['value_to_check']

    __migrations__ = {
        1: TestModelMigrationDoesNotLeaveCorruptedData.migration_with_wait
    }

    def __init__(self, **config):
        self.value_to_check = None
        super().__init__(**config)



import pytest

from tinybird.ch import HTTPClient
from tinybird.pg import PGService
from tinybird.user import Users
from tinybird.token_scope import scopes

from .conftest import CH_HOST, CH_HTTP_PORT, get_clickhouse_version
from .views.base_test import BaseTest


def create_test_tables(database_server, database):
    def _get_replicated_engine(table_name):
        return "ReplicatedMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name + "','{replica}')"

    def _get_replicated_aggregating_engine(table_name):
        return "ReplicatedAggregatingMergeTree('/clickhouse/tables/{layer}-{shard}/" + table_name + "','{replica}')"

    create_tables = [
        f"""
        CREATE TABLE IF NOT EXISTS {database}.ints ON CLUSTER tinybird (
            c1 Int8, c2 Int16, c3 Int32, c4 Int64,
            c5 UInt8, c6 UInt16, c7 UInt32, c8 UInt64,
            c9 Int128, c10 UInt128,
            c11 Int256, c12 UInt256,
            c13 Float32, c14 Nullable(Float64)
        ) ENGINE = {_get_replicated_engine(f'ints')} PARTITION BY c1 ORDER BY (c1);
        """,
        f"""
        CREATE TABLE IF NOT EXISTS {database}.types ON CLUSTER tinybird (
            c1 Date, c2 DateTime, c3 String, c4 FixedString(5), c5 UUID,
            c6 Enum8('one' = 1, 'two' = 2),
            c7 Enum16('one' = 1, 'two' = 2, 'three' = 3),
            c9 Nullable(FixedString(50)),
            c8 LowCardinality(String), c10 Nullable(Bool)
        ) ENGINE = {_get_replicated_engine(f'types')} PARTITION BY c1 ORDER BY (c1);
        """,
        f"""
        CREATE TABLE IF NOT EXISTS {database}.arrays ON CLUSTER tinybird (
            c1 Array(Int),
            c2 Array(String),
            c3 Array(DateTime),
            c4 Array(Tuple(String, Float32))
        ) ENGINE = {_get_replicated_engine(f'arrays')} PARTITION BY c1 ORDER BY (c1);
        """,
        f"""
        CREATE TABLE IF NOT EXISTS {database}.tuples ON CLUSTER tinybird (
            c1 Int8,
            c2 Tuple(Int, String, Float32),
            c3 Nested(a Int, b Int),
            c4 Tuple(Float64, Float64)
        ) ENGINE = {_get_replicated_engine(f'tuples')} PARTITION BY c1 ORDER BY (c1);
        """,
        f"""
        CREATE TABLE IF NOT EXISTS {database}.dt ON CLUSTER tinybird (
            c1 Int32,
            c2 DateTime('Europe/Moscow'),
            c3 DateTime('Europe/Berlin'),
            c4 DateTime
        ) ENGINE = {_get_replicated_engine(f'dt')} PARTITION BY c1 ORDER BY (c1);
        """,
        f"""
        CREATE TABLE IF NOT EXISTS {database}.dt64 ON CLUSTER tinybird (
            c1 Int32,
            c2 DateTime64(6, 'Europe/Moscow'),
            c3 DateTime64(9, 'Europe/Berlin'),
            c4 DateTime64(3),
            c5 DateTime64(6),
            c6 Array(DateTime64(3))
        ) ENGINE = {_get_replicated_engine(f'dt64')} PARTITION BY c1 ORDER BY (c1);
        """,
        f"""
        create table if not exists {database}.t1 on cluster tinybird (a int, b int)
            engine = {_get_replicated_engine(f't1')}
            order by a
        """,
        f"""
        create table if not exists {database}.t2 on cluster tinybird (a int, b AggregateFunction(sum, Int32))
            engine = {_get_replicated_aggregating_engine(f't2')}
            order by a
        """,
        f"""
        create table if not exists {database}.t3 on cluster tinybird (a int, b Array(Int32), c Array(Int32))
            engine = {_get_replicated_engine(f't3')}
            order by a
        """,
        f"""
        create table if not exists {database}.t4 on cluster tinybird (a int,
            b AggregateFunction(sum, Int32),
            c AggregateFunction(sumMap, Array(Int32), Array(Int32)),
            d SimpleAggregateFunction(sum, Int64))
        engine = {_get_replicated_aggregating_engine(f't4')}
        order by a
        """,
        f"""
        create materialized view if not exists {database}.t1_aggr on cluster tinybird
            engine={_get_replicated_aggregating_engine(f't1_aggr')}
            order by a populate as select a, sumState(b) as b from t1 group by a;
        """,
        # """
        # create materialized view t3_aggr
        #     engine=AggregatingMergeTree()
        #     order by a populate as select a, sumMapState(b, c) as b from t3 group by a;
        # """,
        f"""
        CREATE TABLE IF NOT EXISTS {database}.ips ON CLUSTER tinybird (
            c1 IPv4, c2 IPv6
        ) ENGINE = {_get_replicated_engine(f'ips')} PARTITION BY c1 ORDER BY (c1);
        """,
        f"""
        CREATE VIEW IF NOT EXISTS {database}.view ON CLUSTER tinybird (
            c1 Int8, c100 Nullable(Nothing)
        ) AS SELECT c1 FROM {database}.ints;
        """
    ]

    client = HTTPClient(database_server, database=database)
    for query in create_tables:
        client.query_sync(query, read_only=False)

    expect_tables = {
        'ints': f"""CREATE FOREIGN TABLE IF NOT EXISTS "ints" ("c1" INT2 NOT NULL, "c2" INT2 NOT NULL, "c3" INT4 NOT NULL, "c4" INT8 NOT NULL, "c5" INT2 NOT NULL, "c6" INT4 NOT NULL, "c7" INT8 NOT NULL, "c8" NUMERIC NOT NULL, "c9" NUMERIC NOT NULL, "c10" NUMERIC NOT NULL, "c11" NUMERIC NOT NULL, "c12" NUMERIC NOT NULL, "c13" REAL NOT NULL, "c14" DOUBLE PRECISION) SERVER "fdw_{database}" OPTIONS(table_name \'ints\', engine \'ReplicatedMergeTree\')""",
        'types': f"""CREATE FOREIGN TABLE IF NOT EXISTS "types" ("c1" DATE NOT NULL, "c2" TIMESTAMP NOT NULL, "c3" TEXT NOT NULL, "c4" VARCHAR(5) NOT NULL, "c5" UUID NOT NULL, "c6" TEXT NOT NULL, "c7" TEXT NOT NULL, "c9" VARCHAR(50) , "c8" TEXT NOT NULL, "c10" BOOLEAN) SERVER "fdw_{database}" OPTIONS(table_name 'types', engine 'ReplicatedMergeTree')""",
        'arrays': f"""CREATE FOREIGN TABLE IF NOT EXISTS "arrays" ("c1" INT4[] NOT NULL, "c2" TEXT[] NOT NULL, "c3" TIMESTAMP[] NOT NULL, "c4" TEXT[] NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name \'arrays\', engine \'ReplicatedMergeTree\')""",
        'tuples': f"""CREATE FOREIGN TABLE IF NOT EXISTS "tuples" ("c1" INT2 NOT NULL, "c2" TEXT NOT NULL, "c3.a" INT4[] NOT NULL, "c3.b" INT4[] NOT NULL, "c4" TEXT NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name \'tuples\', engine \'ReplicatedMergeTree\')""",
        'dt': f"""CREATE FOREIGN TABLE IF NOT EXISTS "dt" ("c1" INT4 NOT NULL, "c2" TIMESTAMP NOT NULL, "c3" TIMESTAMP NOT NULL, "c4" TIMESTAMP NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name \'dt\', engine \'ReplicatedMergeTree\')""",
        'dt64': f"""CREATE FOREIGN TABLE IF NOT EXISTS "dt64" ("c1" INT4 NOT NULL, "c2" TIMESTAMP NOT NULL, "c3" TIMESTAMP NOT NULL, "c4" TIMESTAMP NOT NULL, "c5" TIMESTAMP NOT NULL, "c6" TIMESTAMP[] NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name \'dt64\', engine \'ReplicatedMergeTree\')""",
        't1': f"""CREATE FOREIGN TABLE IF NOT EXISTS "t1" ("a" INT4 NOT NULL, "b" INT4 NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name 't1', engine 'ReplicatedMergeTree')""",
        't2': f"""CREATE FOREIGN TABLE IF NOT EXISTS "t2" ("a" INT4 NOT NULL, "b" INT4 OPTIONS(AggregateFunction 'sum') NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name 't2', engine 'ReplicatedAggregatingMergeTree')""",
        't3': f"""CREATE FOREIGN TABLE IF NOT EXISTS "t3" ("a" INT4 NOT NULL, "b" INT4[] NOT NULL, "c" INT4[] NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name 't3', engine 'ReplicatedMergeTree')""",
        't4': f"""CREATE FOREIGN TABLE IF NOT EXISTS "t4" ("a" INT4 NOT NULL, "b" INT4 OPTIONS(AggregateFunction 'sum') NOT NULL, "c" TEXT OPTIONS(AggregateFunction 'sumMap') NOT NULL, "d" INT8 OPTIONS(AggregateFunction 'sum') NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name 't4', engine 'ReplicatedAggregatingMergeTree')""",
        't1_aggr': f"""CREATE FOREIGN TABLE IF NOT EXISTS "t1_aggr" ("a" INT4 NOT NULL, "b" INT4 OPTIONS(AggregateFunction 'sum') NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name 't1_aggr', engine 'MaterializedView')""",
        # 't3_aggr': """
        #     create materialized view t3_aggr
        #         engine=AggregatingMergeTree()
        #         order by a populate as select a, sumMapState(b, c) as b from t3 group by a;
        #     """,
        'ips': f"""CREATE FOREIGN TABLE IF NOT EXISTS "ips" ("c1" inet NOT NULL, "c2" inet NOT NULL) SERVER "fdw_{database}" OPTIONS(table_name 'ips', engine 'ReplicatedMergeTree')""",
        'view': f"""CREATE FOREIGN TABLE IF NOT EXISTS "view" ("c1" INT2 NOT NULL, "c100" TEXT) SERVER "fdw_{database}" OPTIONS(table_name 'view', engine 'None')""",  # The engine is View. We do not allow Views inside the product. This is only needed for a ITX "hackerismo".
    }

    return expect_tables


class TestPGService(BaseTest):
    def setUp(self):
        super().setUp()
        u = Users.get_by_id(self.WORKSPACE_ID)
        self.u = u
        self.admin_token = Users.get_token_for_scope(u, scopes.ADMIN)
        # drop the database in case it already exists
        self.u['enabled_pg'] = True
        self.u['pg_server'] = '127.0.0.1'
        self.u['pg_foreign_server'] = CH_HOST
        self.u['pg_foreign_server_port'] = CH_HTTP_PORT
        self.pg_service = PGService(self.u)
        self.pg_service.drop_database()

    def tearDown(self):
        # drop the database in case it already exists
        self.u['enabled_pg'] = True
        self.pg_service.drop_database()
        self.u['enabled_pg'] = False
        super().tearDown()

    def test_does_nothing_if_not_enabled_pg(self):
        self.u['enabled_pg'] = True
        self.pg_service.drop_database()
        self.u['enabled_pg'] = False
        self.pg_service.setup_database()

        result = self.pg_service.execute("SELECT oid FROM pg_database WHERE datname = %s;", params=[self.pg_service.get_database_name()])
        self.assertEqual(len(result), 0)

    def test_setup_database(self):
        self.pg_service.setup_database()
        result = self.pg_service.execute("SELECT oid FROM pg_database WHERE datname = %s;", params=[self.pg_service.get_database_name()])
        self.assertEqual(len(result), 1)

    def test_sync_foreign_tables(self):
        self.pg_service.setup_database()
        result = self.pg_service.execute("SELECT oid FROM pg_database WHERE datname = %s;", params=[self.pg_service.get_database_name()])
        self.assertEqual(len(result), 1)

        self.pg_service.sync_foreign_tables()

        sql = "SELECT * FROM pg_views WHERE viewname = 'test_table' AND schemaname = 'public';"
        result = self.pg_service.execute(sql, role='user')
        self.assertEqual(len(result), 1)

        sql = "SELECT count(*) FROM test_table;"
        result = self.pg_service.execute(sql, role='user')
        self.assertEqual(result[0]['count'], 6)

    @pytest.mark.skipif(get_clickhouse_version() < 22.8, reason="Bug in 21.9 when creating a tablewith ReplicatedAggregatingMergeTree on cluster")
    def test_create_foreign_tables(self):
        self.pg_service.setup_database()
        result = self.pg_service.execute("SELECT oid FROM pg_database WHERE datname = %s;", params=[self.pg_service.get_database_name()])
        self.assertEqual(len(result), 1)

        expect_tables = create_test_tables(self.u['database_server'], self.u['database'])

        for table_name in expect_tables.keys():
            view_name = f'v_{table_name}'

            # pre-checks
            sql = f"SELECT * FROM information_schema.tables WHERE table_type = 'FOREIGN' AND table_name = '{table_name}' AND table_schema = 'public';"
            result = self.pg_service.execute(sql, role='user')
            self.assertEqual(len(result), 0)

            sql = f"SELECT * FROM pg_views WHERE viewname = '{view_name}' AND schemaname = 'public';"
            result = self.pg_service.execute(sql, role='user')
            self.assertEqual(len(result), 0)

            # create
            ft = self.pg_service.create_foreign_table(table_name, view_name)
            expect_sql = ft.replace('\n', ' ').replace('  ', '').strip()
            self.assertEqual(expect_sql, expect_tables[table_name])

            # post-checks
            sql = f"SELECT * FROM information_schema.tables WHERE table_type = 'FOREIGN' AND table_name = '{table_name}' AND table_schema = 'public';"
            result = self.pg_service.execute(sql, role='user')
            self.assertEqual(len(result), 1)

            sql = f"SELECT * FROM pg_views WHERE viewname = '{view_name}' AND schemaname = 'public';"
            result = self.pg_service.execute(sql, role='user')
            self.assertEqual(len(result), 1)

    def test_drop_foreign_table(self):
        self.pg_service.setup_database()
        result = self.pg_service.execute("SELECT oid FROM pg_database WHERE datname = %s;", params=[self.pg_service.get_database_name()])
        self.assertEqual(len(result), 1)

        self.pg_service.sync_foreign_tables()

        self.pg_service.drop_foreign_table(self.u.get_datasources()[0].id)
        sql = "SELECT * FROM pg_views WHERE viewname = 'test_table' AND schemaname = 'public';"
        result = self.pg_service.execute(sql, role='user')
        self.assertEqual(len(result), 0)

    def test_alter_datasource_name(self):
        self.pg_service.setup_database()
        result = self.pg_service.execute("SELECT oid FROM pg_database WHERE datname = %s;", params=[self.pg_service.get_database_name()])
        self.assertEqual(len(result), 1)

        self.pg_service.sync_foreign_tables()

        self.pg_service.alter_datasource_name(self.u.get_datasources()[0].name, 'prapra')
        sql = "SELECT * FROM pg_views WHERE viewname = 'test_table' AND schemaname = 'public';"
        result = self.pg_service.execute(sql, role='user')
        self.assertEqual(len(result), 0)

        sql = "SELECT * FROM pg_views WHERE viewname = 'prapra' AND schemaname = 'public';"
        result = self.pg_service.execute(sql, role='user')
        self.assertEqual(len(result), 1)

        sql = "SELECT count(*) as c FROM prapra;"
        result = self.pg_service.execute(sql, role='user')
        self.assertEqual(result[0]['c'], 6)

    def test_queries_from_postgres_carries_correct_current_database_and_user_agent(self):
        self.pg_service.setup_database()
        result = self.pg_service.execute("SELECT oid FROM pg_database WHERE datname = %s;", params=[self.pg_service.get_database_name()])
        self.assertEqual(len(result), 1)

        self.pg_service.sync_foreign_tables()

        sql = "SELECT count(*) from test_table"
        result = self.pg_service.execute(sql, role='user')
        self.assertEqual(len(result), 1)

        query_log_row_new = self.get_query_logs_by_where(f"query = 'SELECT count(*) FROM {self.u.database}.{self.u.get_datasources()[0].id}'")

        self.assertEqual(query_log_row_new[0]['http_user_agent'], 'postgres')
        self.assertEqual(query_log_row_new[0]['current_database'], self.u.database)



import unittest

from chtoolset import query as chquery
from tinybird.sql_toolset import replace_tables


class TestSQLBasic(unittest.TestCase):
    sql = "SELECT * FROM table_a"

    def test_replace(self):
        replaced = replace_tables(self.sql, {
            'table_a': '__t_id_1'
        })
        self.assertEqual(replaced, chquery.format("SELECT * FROM __t_id_1 as table_a"))

    def test_replace_default_database(self):
        replaced = replace_tables(self.sql, {
            'table_a': '__t_id_1'
        }, default_database='d_012345')
        self.assertEqual(replaced, chquery.format("SELECT * FROM d_012345.__t_id_1 as table_a"))

    def test_replace_tuple_default_database(self):
        replaced = replace_tables("SELECT * FROM db_1.table_a", {
            ('db_1', 'table_a'): '__t_id_1'
        }, default_database='d_012345')
        self.assertEqual(replaced, chquery.format("SELECT * FROM d_012345.__t_id_1 as table_a"))


class TestSQLMultipleTables(unittest.TestCase):
    def test_replace_tables(self):
        sql = """
            SELECT
                table_a.col0,
                a.col1,
                b.col0
            FROM table_a as a any inner join (
                SELECT * FROM db_2.table_b
            ) as b USING unified
        """
        replaced = replace_tables(sql, {
            'table_a': '__t_id_1',
            ('db_2', 'table_b'): '__t_id_2'
        })
        self.assertEqual(replaced, chquery.format("""
            SELECT
                __t_id_1.col0,
                a.col1,
                b.col0
            FROM __t_id_1 as a any inner join (
                SELECT * FROM __t_id_2 as table_b
            ) as b USING unified
        """))

    def test_replace_tables_default_database(self):
        sql = """
            SELECT
                db_1.table_a.*,
                table_a.col0,
                a.col1,
                b.col0
            FROM table_a as a any inner join (
                SELECT * FROM db_2.table_b
            ) as b USING unified
        """
        replaced = replace_tables(sql, {
            'table_a': '__t_id_1',
            ('db_1', 'table_a'): '__t_id_1',
            ('db_2', 'table_b'): '__t_id_2'
        }, default_database='db_1')
        self.assertEqual(replaced, chquery.format("""
            SELECT
                db_1.__t_id_1.*,
                __t_id_1.col0,
                a.col1,
                b.col0
            FROM db_1.__t_id_1 as a any inner join (
                SELECT * FROM db_1.__t_id_2 AS table_b
            ) as b USING unified
        """))


class TestSQLComplexQueries(unittest.TestCase):
    def test_replace_with_join_default_database(self):
        sql = """
        select * from (
            select * from (
                select * from listings where city like '_evill_'
            ) as f
            UNION ALL
            select * from (
                select * from checkings as cc any left join (
                    select unifiedid, city, bedrooms from listings where city = 'Seville'
                ) aa using unifiedid
            ) as l_13
        ) limit 100
        """
        replaced = replace_tables(sql, {
            'listings': 'l1',
            'checkings': 'c1'
        }, default_database='d_012345')
        self.assertEqual(replaced, chquery.format("""
        select * from (
            select * from (
                select * from d_012345.l1 as listings where city like '_evill_'
            ) as f
            UNION ALL
            select * from (
                select * from d_012345.c1 as cc any left join (
                    select unifiedid, city, bedrooms from d_012345.l1 as listings where city = 'Seville'
                ) aa using unifiedid
            ) as l_13
        ) limit 100
        """))

    # Worked well in 0.14.dev1 and previous releases, and 0.15.dev1+.
    # Broken in 0.15.dev0: aliases column names didn't work in the second replacement (the first one was ok)
    # Got:          traceback,
    # Expected:     JSONExtract(tags, 'traceback', 'Array(String)') AS traceback,
    def test_replace_with_join_with_inner_aliases(self):
        self.maxDiff = None
        spans_500_errors_sql = """
            SELECT
                start_datetime,
                workspace_name,
                status_code,
                method,
                url,
                error,
                JSONExtract(tags, 'traceback', 'Array(String)') traceback,
                logs,
                tags,
                operation_name
            FROM Internal.spans
            where start_datetime > now() - INTERVAL 1 day
            and status_code >= 500
            and kind = 'server'
            and component = 'tornado'
            ORDER BY start_datetime DESC
        """

        spans_500_error_count = """
            SELECT
                count() as 500_errors_count,
                500_errors_count as k_500_errors_count,
                uniq(cityHash64(error, traceback)) as uniq_errors
            FROM (
                SELECT * FROM untitled_pipe_7234_0
                WHERE start_datetime > now() - INTERVAL 1 day
            ) a
            join (
                SELECT * FROM untitled_pipe_7234_0
                WHERE start_datetime > now() - INTERVAL 1 day
            ) b
            using k_500_errors_count
            WHERE 1
            ORDER BY 500_errors_count DESC
        """

        replacement = {('default', 'untitled_pipe_7234_0'): ('', '(' + spans_500_errors_sql + ')'), ('Internal', 'spans'): ('', 'spans')}
        replaced = replace_tables(spans_500_error_count, replacement, default_database='default')

        self.assertEqual(replaced, chquery.format("""
            SELECT
                count() AS `500_errors_count`,
                `500_errors_count` AS k_500_errors_count,
                uniq(cityHash64(error, traceback)) AS uniq_errors
            FROM
            (
                SELECT *
                FROM
                (
                    SELECT
                        start_datetime,
                        workspace_name,
                        status_code,
                        method,
                        url,
                        error,
                        JSONExtract(tags, 'traceback', 'Array(String)') AS traceback,
                        logs,
                        tags,
                        operation_name
                    FROM default.spans
                    WHERE (start_datetime > (now() - toIntervalDay(1))) AND (status_code >= 500) AND (kind = 'server') AND (component = 'tornado')
                    ORDER BY start_datetime DESC
                ) AS untitled_pipe_7234_0
                WHERE start_datetime > (now() - toIntervalDay(1))
            ) AS a
            INNER JOIN
            (
                SELECT *
                FROM
                (
                    SELECT
                        start_datetime,
                        workspace_name,
                        status_code,
                        method,
                        url,
                        error,
                        JSONExtract(tags, 'traceback', 'Array(String)') AS traceback,
                        logs,
                        tags,
                        operation_name
                    FROM default.spans
                    WHERE (start_datetime > (now() - toIntervalDay(1))) AND (status_code >= 500) AND (kind = 'server') AND (component = 'tornado')
                    ORDER BY start_datetime DESC
                ) AS untitled_pipe_7234_0
                WHERE start_datetime > (now() - toIntervalDay(1))
            ) AS b USING (k_500_errors_count)
            WHERE 1
            ORDER BY `500_errors_count` DESC
        """))


class TestSQLReplaceTables(unittest.TestCase):

    scenarios = [
        ('select * from tt', {'tt': 'table2'}, 'select * from table2 as tt'),
        ('select * from tt', {'tt': '(select a from table2 where a > 1)'}, 'select * from (select a from table2 where a > 1) AS tt'),
        ('select tt.a from tt', {'tt': '(select a from table2 where a > 1)'}, 'select tt.a from (select a from table2 where a > 1) AS tt'),
        ('select tt.* from tt', {'tt': '(select a from table2 where a > 1)'}, 'select tt.* from (select a from table2 where a > 1) AS tt'),
        ('select aa.a from tt AS aa', {'tt': '(select a from table2 where a > 1)'}, 'select aa.a from (select a from table2 where a > 1) AS aa'),
        ('select * from database.tt', {('database', 'tt'): '(select a from table2 where a > 1)'}, 'select * from (select a from table2 where a > 1) AS tt'),
        ('select * from database.`tt`', {('database', 'tt'): '(select a from table2 where a > 1)'}, 'select * from (select a from table2 where a > 1) AS tt'),
        ('select * from `database`.`tt`', {('database', 'tt'): '(select a from table2 where a > 1)'}, 'select * from (select a from table2 where a > 1) AS tt'),
        ('select * from tt inner join ttj using b', {'tt': 'tt2', 'ttj': 'ttj2'}, 'select * from tt2 as tt inner join ttj2 as ttj using b'),
        ('select * from tt inner join tt using b', {'tt': 'tt2'}, 'select * from tt2 as tt inner join tt2 as tt using b'),
        ('select count() c from test_table format JSON', {'test_table': 'pepe'}, 'select count() c from pepe as test_table format JSON'),
        ('select * from tt', {'tt': ('d_012345', 'table2')}, 'select * from d_012345.table2 as tt'),
        (
            'select count() as t, t.record, avg(landing.timestamp) from landing t group by t.record',
            {'landing': ('database', 't_01010101')},
            'select count() as t, t.record, avg(t_01010101.timestamp) from database.t_01010101 t group by t.record'
        ),
    ]

    def test_replace(self):
        for (sql, replacements, expected_query) in self.scenarios:
            with self.subTest(sql=sql):
                self.assertEqual(replace_tables(sql, replacements), chquery.format(expected_query))


class TestSQLReplaceTablesWithDeps(unittest.TestCase):

    scenarios = [
        (
            'select * from test',
            {
                'test': 'testing'
            },
            'select * from testing AS test'
        ),
        (
            'select * from test',
            {
                'test': '(select * from testing)'
            },
            'select * from (select * from testing) as test'
        ),
        (
            'select * from test',
            {
                'test': '(select * from test0)',
                'test0': 'test_id'
            },
            'select * from (select * from test_id as test0) as test'
        ),
        (
            'select * from test',
            {
                'test': '(select * from test0)',
                'test0': 'test_id',
                'test_id': 'test_id_on_disk'
            },
            'select * from (select * from test_id_on_disk as test0) as test'
        ),
        (
            'select * from test',
            {
                'test': '(select * from test0)',
                'test0': 'test_id',
                'test_id': 'test_id_on_disk',
                'test_id_on_disk': '(select * from test_id_on_disk where a > 10)'
            },
            'select * from (select * from (select * from test_id_on_disk where a > 10) as test0) AS test'
        ),
        (
            'select * from test',
            {
                'test': '(select * from test0)',
                'test0': '(select * from test_id inner join wadus using b)',
                'test_id': 'test_id_on_disk',
                'wadus': '(select * from wadus_src where c < 1000)'
            },
            'select * from (select * from (select * from test_id_on_disk as test_id inner join (select * from wadus_src where c < 1000) as wadus using b) as test0) as test'
        ),
        (
            'select * from test',
            {
                'test': '(select * from test0)',
                'test0': '(select * from test_id inner join wadus using b)',
                'test_id': 'test_id_on_disk',
                'wadus': '(select * from wadus_src where c < 1000)',
                'test_id_on_disk': '(select * from test_id_on_disk where a > 10)'
            },
            'select * from (select * from (select * from (select * from test_id_on_disk where a > 10) as test_id inner join (select * from wadus_src where c < 1000) as wadus using b) as test0) as test'
        ),
        (
            'select * from a left join b on a.id = b.id',
            {
                'a': '(select * from test0 where id > 1)',
                'b': '(select * from test1 where id > 1)',
            },
            'select * from (select * from test0 where id > 1) as a left join (select * from test1 where id > 1) as b on a.id = b.id'
        ),
        (
            'select * from a as a_alias left join b  b_alias on a_alias.id = b_alias.id',
            {
                'a': '(select * from test0 where id > 1)',
                'b': '(select * from test1 where id > 1)',
            },
            'select * from (select * from test0 where id > 1) as a_alias left join (select * from test1 where id > 1) as b_alias on a_alias.id = b_alias.id'
        ),
        # https://gitlab.com/tinybird/analytics/-/issues/1964
        (
            'SELECT count() FROM test_pipe_1',
            {
                'test_pipe_1': '(SELECT distinct n FROM node0)',
                'node0': '(SELECT number % 2 n from numbers(100))',
            },
            'SELECT count() FROM (SELECT distinct n FROM (SELECT number % 2 n from numbers(100)) as node0) as test_pipe_1'
        )
    ]

    def test_replace(self):
        for (sql, replacements, expected_query) in self.scenarios:
            with self.subTest(sql=sql, replacements=replacements):
                self.assertEqual(replace_tables(sql, replacements), chquery.format(expected_query))

    def test_cycle(self):
        sql = 'select * from test'
        replacements = {
            'test': '(select * from test0)',
            'test0': '(select * from test_id inner join wadus using b)',
            'test_id': 'test_id_on_disk',
            'wadus': '(select * from foo where c < 1000)',  # <-----|
            'foo': '(select * from wadus)',  # cyclic dependency ---|
            'test_id_on_disk': '(select * from test_id_on_disk where a > 10)'
        }
        with self.assertRaisesRegex(ValueError, 'Circular dependencies exist'):
            replace_tables(sql, replacements)

    def test_other_database(self):
        sql = 'select * from nyc_taxi'
        replacements = {
            'nyc_taxi': '(select * from public.nyc_taxi)'
        }

        expected_query = 'select * from (select * from public.nyc_taxi) as nyc_taxi'

        replaced = replace_tables(sql, replacements)
        self.assertEqual(replaced, chquery.format(expected_query))

    def test_other_database_with_tuple(self):
        sql = 'select * from nyc_taxi'
        replacements = {
            'nyc_taxi': ('public', 'nyc_taxi')
        }

        expected_query = 'select * from public.nyc_taxi'

        replaced = replace_tables(sql, replacements)
        self.assertEqual(replaced, chquery.format(expected_query))

    def test_join_get(self):
        replacements = {
            'abcd': 't_id_abcd'
        }
        expected_query = "select joinGet('t_id_abcd', 'test', 1)"
        replaced = replace_tables("select joinGet('abcd', 'test', 1)", replacements)
        self.assertEqual(replaced, chquery.format(expected_query))

        expected_query = "select joinGet(t_id_abcd, 'test', 1)"
        replaced = replace_tables("select joinGet(abcd, 'test', 1)", replacements)
        self.assertEqual(replaced, chquery.format(expected_query))

    def test_join_get_tuple(self):
        replacements = {
            'abcd': ('db_1', 't_id_abcd')
        }

        expected_query = "select joinGet('db_1.t_id_abcd', 'test', 1)"
        replaced = replace_tables("select joinGet('abcd', 'test', 1)", replacements)
        self.assertEqual(replaced, chquery.format(expected_query))

        expected_query = "select joinGet(db_1.t_id_abcd, 'test', 1)"
        replaced = replace_tables("select joinGet(abcd, 'test', 1)", replacements)
        self.assertEqual(replaced, chquery.format(expected_query))

    def test_join_get_default_database(self):
        replacements = {
            'abcd': 't_id_abcd'
        }

        expected_query = "select joinGet('d_012345.t_id_abcd', 'test', 1)"
        replaced = replace_tables("select joinGet('abcd', 'test', 1)", replacements, default_database='d_012345')
        self.assertEqual(replaced, chquery.format(expected_query))

        expected_query = "select joinGet(d_012345.t_id_abcd, 'test', 1)"
        replaced = replace_tables("select joinGet(abcd, 'test', 1)", replacements, default_database='d_012345')
        self.assertEqual(replaced, chquery.format(expected_query))

    def test_join_get_nested(self):
        replacements = {
            't1': 't_id_1',
            't2': 't_id_2',
            't3': 't_id_3',
        }
        expected_query = "SELECT joinGet('d_012345.t_id_1', 'c1', concat(joinGet('d_012345.t_id_2', 'c2', i2), '_', joinGet('d_012345.t_id_3', 'c3', i3), '_', 'label'))"
        replaced = replace_tables("SELECT joinGet('t1', 'c1', concat(joinGet('t2', 'c2', i2), '_', joinGet('t3', 'c3', i3), '_', 'label'))", replacements, default_database='d_012345')
        self.assertEqual(replaced, chquery.format(expected_query))

    def test_join_nested_2(self):
        replacements = {
            'blabla': 'alibaba',
            'pp': 't_pp_01'
        }
        query = "select joinGet(blabla, 'a', joinGet(pp, 'b', c))"
        replaced = replace_tables(query, replacements, default_database='d_012345')
        expected_query = "select joinGet(d_012345.alibaba, 'a', joinGet(d_012345.t_pp_01, 'b', c))"
        self.assertEqual(replaced, chquery.format(expected_query))

    def test_invalid_subquery(self):
        sql = """SELECT * FROM clients"""
        replacements = {
            'clients': "(select * from clients where country 'ES')",
        }

        with self.assertRaisesRegex(ValueError, r'Syntax error'):
            replace_tables(sql, replacements)



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional
import click
from humanfriendly.tables import format_pretty_table
from tinybird.client import TinyB
import yaml

from tinybird.feedback_manager import FeedbackManager


@ dataclass
class TestCase:
    name: str
    sql: str
    max_time: Optional[float]
    max_bytes_read: Optional[int]

    def __init__(self, name, sql, max_time: Optional[float] = None, max_bytes_read: Optional[int] = None):
        self.name = name
        self.sql = sql
        self.max_time = max_time
        self.max_bytes_read = max_bytes_read

    def __iter__(self):
        yield (self.name, {'sql': self.sql, 'max_time': self.max_time, 'max_bytes_read': self.max_bytes_read})


@ dataclass(frozen=True)
class Status():
    name: str
    abbreviation: str
    color: str
    description: str


PASS = Status(name='PASS', abbreviation='P', description='Pass', color='green')
PASS_OVER_TIME = Status(name='PASS_OVER_TIME', abbreviation='P*OT', description='Pass Over Time', color='cyan')
PASS_OVER_READ_BYTES = Status(name='PASS_OVER_READ_BYTES', abbreviation='P*OB', description='Pass Over Read Bytes', color='cyan')
PASS_OVER_TIME_AND_READ_BYTES = Status(name='PASS_OVER_TIME_AND_OVER_BYTES', abbreviation='P*OT*OB', description='Pass Over Time and Over Read Bytes', color='cyan')
FAILED = Status(name='FAIL', abbreviation='F', description='Fail', color='red')
ERROR = Status(name='ERROR', abbreviation='E', description='Error', color='bright_yellow')


@ dataclass()
class TestResult:
    name: str
    data: List[Dict]
    elapsed_time: float
    read_bytes: int
    max_elapsed_time: Optional[float]
    max_bytes_read: Optional[int]
    error: Optional[str]

    @ property
    def status(self) -> Status:
        if len(self.data) > 0:
            return FAILED
        elif self.max_bytes_read is not None and self.max_elapsed_time is not None and self.read_bytes > self.max_bytes_read and self.elapsed_time > self.max_elapsed_time:
            return PASS_OVER_TIME_AND_READ_BYTES
        elif self.max_bytes_read is not None and self.read_bytes > self.max_bytes_read:
            return PASS_OVER_READ_BYTES
        elif self.max_elapsed_time is not None and self.elapsed_time > self.max_elapsed_time:
            return PASS_OVER_TIME
        elif self.error:
            return ERROR
        return PASS

    def __dict__(self):
        return {
            'name': self.name,
            'data': self.data,
            'elapsed_time': self.elapsed_time,
            'read_bytes': self.read_bytes,
            'max_elapsed_time': self.max_elapsed_time,
            'max_bytes_read': self.max_bytes_read,
            'error': self.error,
            'status': self.status.name
        }


@dataclass()
class TestSummaryResults:
    filename: str
    results: List[TestResult]


def parse_file(file: str) -> Iterable[TestCase]:
    with Path(file).open('r') as f:
        definitions: List[Dict[str, Any]] = yaml.safe_load(f)

    for definition in definitions:
        try:
            for name, properties in definition.items():
                yield TestCase(
                    name,
                    properties.get('sql'),
                    properties.get('max_time'),
                    properties.get('max_bytes_read'))
        except Exception as e:
            click.echo(f"""Error: {FeedbackManager.error_exception(error=e)} reading file, check "{file}"->"{definition.get('name')}" """)


def generate_file(file: str, overwrite: bool = False) -> None:
    definitions = [
        dict(TestCase('this_test_should_pass', sql='SELECT * FROM numbers(5) WHERE 0')),
        dict(TestCase('this_test_should_fail', 'SELECT * FROM numbers(5) WHERE 1')),
        dict(TestCase('this_test_should_pass_over_time', 'SELECT * FROM numbers(5) WHERE 0', max_time=0.0000001)),
        dict(TestCase('this_test_should_pass_over_bytes', 'SELECT sum(number) AS total FROM numbers(5) HAVING total>1000', max_bytes_read=5)),
        dict(TestCase('this_test_should_pass_over_time_and_bytes', 'SELECT sum(number) AS total FROM numbers(5) HAVING total>1000', max_time=0.0000001, max_bytes_read=5)),
    ]

    p = Path(file)
    if ((not p.exists()) or overwrite):
        p.parent.mkdir(parents=True, exist_ok=True)
        with p.open('w') as f:
            yaml.safe_dump(definitions, f)
        click.echo(FeedbackManager.success_generated_local_file(file=p))
    else:
        click.echo(FeedbackManager.error_file_already_exists(file=p))

    return


async def run_test_file(tb_client: TinyB, file: str) -> List[TestResult]:
    results: List[TestResult] = []
    for test_case in parse_file(file):
        if not test_case.sql:
            results.append(TestResult(
                name=test_case.name,
                data=[],
                elapsed_time=0,
                read_bytes=0,
                max_elapsed_time=test_case.max_time,
                max_bytes_read=test_case.max_bytes_read,
                error="Not sql"
            ))
            continue

        q = f"SELECT * FROM ({test_case.sql}) LIMIT 20 FORMAT JSON"
        try:
            test_response = await tb_client.query(q)
            results.append(TestResult(
                name=test_case.name,
                data=test_response['data'],
                elapsed_time=test_response.get('statistics', {}).get('elapsed', 0),
                read_bytes=test_response.get('statistics', {}).get('bytes_read', 0),
                max_elapsed_time=test_case.max_time,
                max_bytes_read=test_case.max_bytes_read,
                error=None
            ))

        except Exception as e:
            results.append(TestResult(
                name=test_case.name,
                data=[],
                elapsed_time=0,
                read_bytes=0,
                max_elapsed_time=test_case.max_time,
                max_bytes_read=test_case.max_bytes_read,
                error=str(e)
            ))

    return results


def test_run_summary(results: List[TestSummaryResults], only_fail: bool = False, verbose_level: int = 0):
    total_counts: Dict[Status, int] = {}
    for result in results:

        summary: List[Dict] = []
        for test in result.results:
            total_counts[test.status] = total_counts.get(test.status, 0) + 1

            # Skip the PASS tests if we only want the failed ones
            if only_fail and test.status in [PASS]:
                continue

            summary.append({
                'file': result.filename,
                'test': test.name,
                'status': test.status.description,
                'elapsed': f"{test.elapsed_time} ms",
            })

        click.echo(format_pretty_table(
            data=[d.values() for d in summary],
            column_names=list(summary[0].keys()) if len(summary) > 0 else []
        ))
        click.echo('\n')

        # Only display the data for debugging when wanting verbose
        if verbose_level == 0:
            continue

        failed_tests = [test for test in result.results if test.status is not PASS]
        for test in failed_tests:
            click.secho(f'{result.filename}::{test.name}', fg=test.status.color, bold=True, nl=True)

            if test.data:
                click.echo(format_pretty_table(
                    data=[d.values() for d in test.data],
                    column_names=list(test.data[0].keys()) if len(test.data) else []
                ))
                click.echo('\n')

            if test.error:
                click.secho(test.error, fg=test.status.color, bold=True, nl=True, err=True)

    if (len(total_counts)):
        click.echo("\nTotals:")
        for key_status, value_total in total_counts.items():
            code_summary = f"Total {key_status.description}: {value_total}"
            click.secho(code_summary, fg=key_status.color, bold=True, nl=True)

    if total_counts.get(FAILED, 0) > 0:
        raise click.ClickException(FeedbackManager.error_some_data_validation_have_failed())


def get_bare_url(url: str) -> str:
    if url.startswith("http://"):
        return url[7:]
    elif url.startswith("https://"):
        return url[8:]
    else:
        return url
